from ludwig.schema.model_config import ModelConfig
from ludwig.utils.config_utils import has_pretrained_encoder, has_trainable_encoder, has_unstructured_input_feature


def get_auto_learning_rate(config: ModelConfig) -> float:
    """Uses config heuristics to determine an appropriate learning rate.

    The main idea behind the following heuristics is that smaller learning rates are more
    suitable for features with larger encoders, which are typically used with unstructured features.
    Note that these are meant to be rough heuristics that are solely based on feature types and the
    type of the corresponding encoder. More factors could be taken into consideration such as model
    size, dataset size, batch size, number of features, etc.

    Args:
        config: Ludwig config used to train the model.
    """
    if not has_unstructured_input_feature(config):
        return 0.001

    if not has_pretrained_encoder(config):
        return 0.0001

    if has_trainable_encoder(config):
        return 0.00001

    return 0.00002



from dataclasses import field
from typing import Dict, Optional, Union

from marshmallow import fields, ValidationError

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import RAY
from ludwig.schema import utils as schema_utils
from ludwig.schema.hyperopt.scheduler import BaseSchedulerConfig, SchedulerDataclassField
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class ExecutorConfig(schema_utils.BaseMarshmallowConfig):
    """Basic executor settings."""

    type: str = schema_utils.ProtectedString(RAY)

    num_samples: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description=(
            "This parameter, along with the space specifications in the parameters section, controls how many "
            "trials are generated."
        ),
    )

    time_budget_s: int = schema_utils.PositiveInteger(
        default=3600, allow_none=True, description="The number of seconds for the entire hyperopt run."
    )

    trial_driver_resources: Dict[str, float] = schema_utils.Dict(
        default=None,
        description=(
            "The resources reserved by each trial driver. This differs from cpu_resources_per_trial and "
            "gpu_resources_per_trial because these resources are reserved for the driver, not its subsequent "
            "workers. Only used when the trials themselves are on the Ray backend. Defaults to 1 CPU."
        ),
    )

    cpu_resources_per_trial: int = schema_utils.PositiveInteger(
        default=1, description="The number of CPU cores allocated to each trial"
    )

    gpu_resources_per_trial: int = schema_utils.NonNegativeInteger(
        default=0, description="The number of GPU devices allocated to each trial"
    )

    kubernetes_namespace: Optional[str] = schema_utils.String(
        default=None,
        allow_none=True,
        description=(
            "When running on Kubernetes, provide the namespace of the Ray cluster to sync results between "
            "pods. See the Ray docs for more info."
        ),
    )

    max_concurrent_trials: Union[str, int, None] = schema_utils.OneOfOptionsField(
        default="auto",
        allow_none=True,
        description=("The maximum number of trials to train concurrently. Defaults to auto if not specified."),
        field_options=[
            schema_utils.PositiveInteger(
                default=1, allow_none=False, description="Manually set a number of concurrent trials."
            ),
            schema_utils.StringOptions(
                options=["auto"],
                default="auto",
                allow_none=False,
                description="Automatically set number of concurrent trials.",
            ),
        ],
    )

    scheduler: BaseSchedulerConfig = SchedulerDataclassField(description="")


@DeveloperAPI
def ExecutorDataclassField(description: str, default: Dict = {}):
    class ExecutorMarshmallowField(fields.Field):
        def _deserialize(self, value, attr, data, **kwargs):
            if isinstance(value, dict):
                try:
                    return ExecutorConfig.Schema().load(value)
                except (TypeError, ValidationError):
                    raise ValidationError(f"Invalid params for executor: {value}, see ExecutorConfig class.")
            raise ValidationError("Field should be dict")

        def _jsonschema_type_mapping(self):
            return {
                **schema_utils.unload_jsonschema_from_marshmallow_class(ExecutorConfig),
                "title": "executor",
                "description": description,
            }

    if not isinstance(default, dict):
        raise ValidationError(f"Invalid default: `{default}`")

    load_default = lambda: ExecutorConfig.Schema().load(default)
    dump_default = ExecutorConfig.Schema().dump(default)

    return field(
        metadata={
            "marshmallow_field": ExecutorMarshmallowField(
                allow_none=False,
                load_default=load_default,
                dump_default=dump_default,
                metadata={"description": description, "parameter_metadata": None},
            )
        },
        default_factory=load_default,
    )



from typing import Tuple

import torch

from ludwig.utils.torch_utils import LudwigModule


def assert_output_shapes(module: LudwigModule, input_shape: Tuple[int]):
    """Runs a unit test to confirm that the out shape matches expected output.

    module: Module to be tested.
    input_shape: List of integers of the expected input shape (w/o batch dim).
    """

    inputs = torch.rand(2, *input_shape, dtype=module.input_dtype)
    output_tensor = module(inputs)
    assert output_tensor.shape[1:] == module.output_shape



from typing import Dict, List, Literal, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import LOGITS, MAXIMIZE, MINIMIZE, PREDICTIONS, PROBABILITIES
from ludwig.utils.registry import Registry

metric_feature_type_registry = Registry()
metric_registry = Registry()
metric_objective_registry = Registry()
metric_tensor_input_registry = Registry()


def register_metric(
    name: str,
    feature_types: Union[str, List[str]],
    objective: Literal[MINIMIZE, MAXIMIZE],
    output_feature_tensor_name: Literal[PREDICTIONS, PROBABILITIES, LOGITS],
):
    """Registers a metric class.

    Args:
        name: The name of the metric. Used in metric reporting and in the config.
        feature_types: The feature types that this metric can be used with.
        objective: The objective of the metric. Either MINIMIZE or MAXIMIZE.
        output_feature_tensor_name: Name of the tensor from output_feature::predictions() that should be used as input.
            For example: PREDICTIONS would be used for accuracy metrics while LOGITS would be used for loss metrics.
    """
    if isinstance(feature_types, str):
        feature_types = [feature_types]

    def wrap(cls):
        for feature_type in feature_types:
            feature_registry = metric_feature_type_registry.get(feature_type, {})
            feature_registry[name] = cls
            metric_feature_type_registry[feature_type] = feature_registry
        metric_registry[name] = cls
        metric_objective_registry[name] = objective
        metric_tensor_input_registry[name] = output_feature_tensor_name
        return cls

    return wrap


def get_metric_classes(feature_type: str):
    return metric_feature_type_registry[feature_type]


def get_metric_cls(feature_type: str, name: str):
    return metric_feature_type_registry[feature_type][name]


@DeveloperAPI
def get_metric_feature_type_registry() -> Registry:
    return metric_feature_type_registry


@DeveloperAPI
def get_metric_registry() -> Registry:
    return metric_registry


@DeveloperAPI
def get_metric(metric_name: str) -> "LudwigMetric":  # noqa
    return get_metric_registry()[metric_name]


@DeveloperAPI
def get_metrics_for_type(feature_type: str) -> Dict[str, "LudwigMetric"]:  # noqa
    return get_metric_feature_type_registry()[feature_type]


@DeveloperAPI
def get_metric_names_for_type(feature_type: str) -> List[str]:
    return sorted(list(get_metric_feature_type_registry()[feature_type].keys()))


@DeveloperAPI
def get_metric_objective(metric_name: str) -> Literal[MINIMIZE, MAXIMIZE]:
    return metric_objective_registry[metric_name]


@DeveloperAPI
def get_metric_tensor_input(metric_name: str) -> Literal[PREDICTIONS, PROBABILITIES, LOGITS]:
    return metric_tensor_input_registry[metric_name]



from typing import Optional

import pytest
import torch

from ludwig.modules.normalization_modules import GhostBatchNormalization

BATCH_SIZE = 16
OUTPUT_SIZE = 8


@pytest.mark.parametrize("virtual_batch_size", [None, BATCH_SIZE // 2, BATCH_SIZE - 14, BATCH_SIZE - 10])
@pytest.mark.parametrize("mode", [True, False])  # training (True) or eval(False)
def test_ghostbatchnormalization(mode: bool, virtual_batch_size: Optional[int]) -> None:
    # setup up GhostBatchNormalization layer
    ghost_batch_norm = GhostBatchNormalization(OUTPUT_SIZE, virtual_batch_size=virtual_batch_size)

    # set training or eval mode
    ghost_batch_norm.train(mode=mode)

    # setup inputs to test
    inputs = torch.randn([BATCH_SIZE, OUTPUT_SIZE], dtype=torch.float32)

    # run tensor through
    norm_tensor = ghost_batch_norm(inputs)

    # check for correctness of output
    assert isinstance(norm_tensor, torch.Tensor)
    assert norm_tensor.shape == (BATCH_SIZE, OUTPUT_SIZE)

    # check for required properties
    assert ghost_batch_norm.input_shape == inputs.shape[1:]
    assert ghost_batch_norm.output_shape == inputs.shape[1:]
    assert ghost_batch_norm.input_dtype == torch.float32

    assert isinstance(ghost_batch_norm.moving_mean, torch.Tensor)
    assert ghost_batch_norm.moving_mean.shape == (OUTPUT_SIZE,)

    assert isinstance(ghost_batch_norm.moving_variance, torch.Tensor)
    assert ghost_batch_norm.moving_variance.shape == (OUTPUT_SIZE,)


def test_ghostbatchnormalization_chunk_size_2() -> None:
    """Test GhostBatchNormalization with virtual_batch_size=2 and batch_size=7 This creates chunks of size 2, 2, 2,
    1 which should be handled correctly since we should skip applying batch norm to the last chunk since it is size
    1."""
    # setup up GhostBatchNormalization layer
    ghost_batch_norm = GhostBatchNormalization(6, virtual_batch_size=2)

    # setup inputs to test
    inputs = torch.randn([7, 6], dtype=torch.float32)

    # Set to training mode
    ghost_batch_norm.train(mode=True)

    # run tensor through
    ghost_batch_norm(inputs)



from typing import List, Tuple, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import MEAN_SQUARED_ERROR, MODEL_ECD, MODEL_GBM, NUMBER
from ludwig.schema import utils as schema_utils
from ludwig.schema.decoders.base import BaseDecoderConfig
from ludwig.schema.decoders.utils import DecoderDataclassField
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import EncoderDataclassField
from ludwig.schema.features.base import BaseInputFeatureConfig, BaseOutputFeatureConfig
from ludwig.schema.features.loss.loss import BaseLossConfig
from ludwig.schema.features.loss.utils import LossDataclassField
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import PreprocessingDataclassField
from ludwig.schema.features.utils import (
    defaults_config_registry,
    ecd_input_config_registry,
    gbm_input_config_registry,
    input_mixin_registry,
    output_config_registry,
    output_mixin_registry,
)
from ludwig.schema.metadata import FEATURE_METADATA
from ludwig.schema.metadata.parameter_metadata import INTERNAL_ONLY
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass


@DeveloperAPI
@input_mixin_registry.register(NUMBER)
@ludwig_dataclass
class NumberInputFeatureConfigMixin(BaseMarshmallowConfig):
    """NumberInputFeatureConfigMixin is a dataclass that configures the parameters used in both the number input
    feature and the number global defaults section of the Ludwig Config."""

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type=NUMBER)


@DeveloperAPI
@ludwig_dataclass
class NumberInputFeatureConfig(BaseInputFeatureConfig, NumberInputFeatureConfigMixin):
    """NumberInputFeatureConfig is a dataclass that configures the parameters used for a number input feature."""

    encoder: BaseEncoderConfig = None


@DeveloperAPI
@ecd_input_config_registry.register(NUMBER)
@ludwig_dataclass
class ECDNumberInputFeatureConfig(NumberInputFeatureConfig):
    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=NUMBER,
        default="passthrough",
    )


@DeveloperAPI
@gbm_input_config_registry.register(NUMBER)
@ludwig_dataclass
class GBMNumberInputFeatureConfig(NumberInputFeatureConfig):
    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_GBM,
        feature_type=NUMBER,
        default="passthrough",
    )


@DeveloperAPI
@output_mixin_registry.register(NUMBER)
@ludwig_dataclass
class NumberOutputFeatureConfigMixin(BaseMarshmallowConfig):
    """NumberOutputFeatureConfigMixin is a dataclass that configures the parameters used in both the number output
    feature and the number global defaults section of the Ludwig Config."""

    decoder: BaseDecoderConfig = DecoderDataclassField(
        feature_type=NUMBER,
        default="regressor",
    )

    loss: BaseLossConfig = LossDataclassField(
        feature_type=NUMBER,
        default=MEAN_SQUARED_ERROR,
    )


@DeveloperAPI
@output_config_registry.register(NUMBER)
@ludwig_dataclass
class NumberOutputFeatureConfig(BaseOutputFeatureConfig, NumberOutputFeatureConfigMixin):
    """NumberOutputFeatureConfig is a dataclass that configures the parameters used for a category output
    feature."""

    clip: Union[List[int], Tuple[int]] = schema_utils.FloatRangeTupleDataclassField(
        n=2,
        default=None,
        allow_none=True,
        min=0,
        max=999999999,
        description="Clip the predicted output to the specified range.",
        parameter_metadata=FEATURE_METADATA[NUMBER]["clip"],
    )

    default_validation_metric: str = schema_utils.StringOptions(
        [MEAN_SQUARED_ERROR],
        default=MEAN_SQUARED_ERROR,
        description="Internal only use parameter: default validation metric for number output feature.",
        parameter_metadata=INTERNAL_ONLY,
    )

    dependencies: list = schema_utils.List(
        default=[],
        description="List of input features that this feature depends on.",
        parameter_metadata=FEATURE_METADATA[NUMBER]["dependencies"],
    )

    reduce_dependencies: str = schema_utils.ReductionOptions(
        default="sum",
        description="How to reduce the dependencies of the output feature.",
        parameter_metadata=FEATURE_METADATA[NUMBER]["reduce_dependencies"],
    )

    reduce_input: str = schema_utils.ReductionOptions(
        default="sum",
        description="How to reduce an input that is not a vector, but a matrix or a higher order tensor, on the first "
        "dimension (second if you count the batch dimension)",
        parameter_metadata=FEATURE_METADATA[NUMBER]["reduce_input"],
    )

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type="number_output")


@DeveloperAPI
@defaults_config_registry.register(NUMBER)
@ludwig_dataclass
class NumberDefaultsConfig(NumberInputFeatureConfigMixin, NumberOutputFeatureConfigMixin):
    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=NUMBER,
        default="passthrough",
    )



import asyncio
import functools
import logging
import os
import shutil
import uuid
from concurrent.futures import ThreadPoolExecutor
from types import ModuleType
from typing import Any, Dict, List, Tuple, Union

import fsspec
import pandas as pd
import yaml

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, CATEGORY
from ludwig.datasets import model_configs_for_dataset
from ludwig.datasets.loaders.dataset_loader import DatasetLoader
from ludwig.globals import CONFIG_YAML
from ludwig.utils.data_utils import load_yaml
from ludwig.utils.dataset_utils import get_repeatable_train_val_test_split
from ludwig.utils.defaults import default_random_seed
from ludwig.utils.fs_utils import get_fs_and_path

HYPEROPT_OUTDIR_RETAINED_FILES = [
    "hyperopt_statistics.json",
    "params.json",
    "stderr",
    "stdout",
    "result.json",
    "error.txt",
]
logger = logging.getLogger()


def load_from_module(
    dataset_module: Union[DatasetLoader, ModuleType], output_feature: Dict[str, str], subsample_frac: float = 1
) -> pd.DataFrame:
    """Load the ludwig dataset, optionally subsamples it, and returns a repeatable split. A stratified split is
    used for classification datasets.

    Args:
        dataset_module: ludwig datasets module (e.g. ludwig.datasets.sst2, ludwig.datasets.ames_housing, etc.)
        subsample_frac: percentage of the total dataset to load.
    """
    dataset = dataset_module.load(split=False)
    if subsample_frac < 1:
        dataset = dataset.sample(frac=subsample_frac, replace=False, random_state=default_random_seed)

    if output_feature["type"] in [CATEGORY, BINARY]:
        return get_repeatable_train_val_test_split(
            dataset,
            stratify_colname=output_feature["name"],
            random_seed=default_random_seed,
        )
    else:
        return get_repeatable_train_val_test_split(dataset, random_seed=default_random_seed)


def export_artifacts(experiment: Dict[str, str], experiment_output_directory: str, export_base_path: str):
    """Save the experiment artifacts to the `bench_export_directory`.

    Args:
        experiment: experiment dict that contains "dataset_name" (e.g. ames_housing),
            "experiment_name" (specified by user), and "config_path" (path to experiment config.
            Relative to ludwig/benchmarks/configs).
        experiment_output_directory: path where the model, data, and logs of the experiment are saved.
        export_base_path: remote or local path (directory) where artifacts are
            exported. (e.g. s3://benchmarking.us-west-2.ludwig.com/bench/ or your/local/bench/)
    """
    protocol, _ = fsspec.core.split_protocol(export_base_path)
    fs, _ = get_fs_and_path(export_base_path)
    try:
        export_full_path = os.path.join(export_base_path, experiment["dataset_name"], experiment["experiment_name"])

        # override previous experiment with the same name
        if fs.exists(export_full_path):
            fs.rm(export_full_path, recursive=True)
        fs.put(experiment_output_directory, export_full_path, recursive=True)
        fs.put(
            os.path.join(experiment["config_path"]),
            os.path.join(export_full_path, CONFIG_YAML),
        )
        logger.info(f"Uploaded experiment artifact to\n\t{export_full_path}")
    except Exception:
        logger.exception(
            f"Failed to upload experiment artifacts for experiment *{experiment['experiment_name']}* on "
            f"dataset {experiment['dataset_name']}"
        )


def download_artifacts(
    bench_config_path: str,
    base_experiment: str,
    experimental_experiment: str,
    download_base_path: str,
    local_dir: str = "benchmarking_summaries",
) -> Tuple[str, List[str]]:
    """Download benchmarking artifacts for two experiments.

    Args:
        bench_config_path: bench config file path. Can be the same one that was used to run
            these experiments.
        base_experiment: name of the experiment we're comparing against.
        experimental_experiment: name of the experiment we're comparing.
        download_base_path: base path under which live the stored artifacts of
            the benchmarking experiments.
    """
    bench_config = load_yaml(bench_config_path)
    protocol, _ = fsspec.core.split_protocol(download_base_path)
    fs, _ = get_fs_and_path(download_base_path)
    os.makedirs(local_dir, exist_ok=True)

    coroutines = []
    for experiment in bench_config["experiments"]:
        dataset_name = experiment["dataset_name"]
        for experiment_name in [base_experiment, experimental_experiment]:
            coroutines.append(download_one(fs, download_base_path, dataset_name, experiment_name, local_dir))
    loop = asyncio.get_event_loop()
    futures = asyncio.gather(*coroutines, return_exceptions=True)
    downloaded_names = loop.run_until_complete(futures)

    dataset_names = [experiment_tuple[0] for experiment_tuple in set(downloaded_names) if experiment_tuple[0]]
    assert (
        len({experiment_tuple[1] for experiment_tuple in downloaded_names}) == 1 and downloaded_names[0][1] == local_dir
    ), "Experiments not downloaded to the same path"

    return local_dir, dataset_names


@DeveloperAPI
async def download_one(
    fs, download_base_path: str, dataset_name: str, experiment_name: str, local_dir: str
) -> Tuple[str, str]:
    """Download `config.yaml` and `report.json` for an experiment.

    Args:
        fs: filesystem to use to download.
        download_base_path: base path under which live the stored artifacts of
            the benchmarking experiments.
        dataset_name: name of the dataset we ran the experiments on.
        experiment_name: name of the experiment (e.g. `v0.5.3_with_bert`)
        local_dir: local directory under which the artifacts will be downloaded.
    """
    loop = asyncio.get_running_loop()
    local_experiment_dir = os.path.join(local_dir, dataset_name, experiment_name)
    remote_experiment_directory = os.path.join(download_base_path, dataset_name, experiment_name)
    os.makedirs(local_experiment_dir, exist_ok=True)
    try:
        with ThreadPoolExecutor() as pool:
            func = functools.partial(
                fs.get,
                remote_experiment_directory,
                local_experiment_dir,
                recursive=True,
            )
            await loop.run_in_executor(pool, func)
    except Exception:
        logger.exception(f"Couldn't download experiment *{experiment_name}* of dataset *{dataset_name}*.")
        return "", local_dir
    return dataset_name, local_dir


def validate_benchmarking_config(benchmarking_config: Dict[str, Any]) -> None:
    """Validates the parameters of the benchmarking config.

    Args:
        benchmarking_config: benchmarking config dictionary.

    Raises:
        ValueError if any of the expected parameters is not there.
    """
    if "experiment_name" not in benchmarking_config and not all(
        "experiment_name" in experiment for experiment in benchmarking_config["experiments"]
    ):
        raise ValueError("You must either specify a global experiment name or an experiment name for each experiment.")
    if "export" not in benchmarking_config:
        raise ValueError(
            """You must specify export parameters. Example:
            export:
              export_artifacts: true
              export_base_path: s3://benchmarking.us-west-2.ludwig.com/bench/    # include the slash at the end.
        """
        )
    if "experiments" not in benchmarking_config:
        raise ValueError("You must specify a list of experiments.")
    for experiment in benchmarking_config["experiments"]:
        if "dataset_name" not in experiment:
            raise ValueError("A Ludwig dataset must be specified.")


def populate_benchmarking_config_with_defaults(benchmarking_config: Dict[str, Any]) -> Dict[str, Any]:
    """Populates the parameters of the benchmarking config with defaults.

    Args:
        benchmarking_config: benchmarking config dictionary.
    """
    if "hyperopt" not in benchmarking_config:
        benchmarking_config["hyperopt"] = False
    if "process_config_file_path" not in benchmarking_config:
        benchmarking_config["process_config_file_path"] = None
    if "profiler" not in benchmarking_config:
        benchmarking_config["profiler"] = {"enable": False, "use_torch_profiler": False, "logging_interval": None}
    return benchmarking_config


def propagate_global_parameters(benchmarking_config: Dict[str, Any]) -> Dict[str, Any]:
    """Propagate the global parameters of the benchmarking config to local experiments.

    Args:
        benchmarking_config: benchmarking config dictionary.
    """
    for experiment in benchmarking_config["experiments"]:
        if "experiment_name" not in experiment:
            experiment["experiment_name"] = benchmarking_config["experiment_name"]
        if "export" not in experiment:
            experiment["export"] = benchmarking_config["export"]
        if "hyperopt" not in experiment:
            experiment["hyperopt"] = benchmarking_config["hyperopt"]
        if "process_config_file_path" not in experiment:
            experiment["process_config_file_path"] = benchmarking_config["process_config_file_path"]
        if "profiler" not in experiment:
            experiment["profiler"] = benchmarking_config["profiler"]
    return benchmarking_config


def create_default_config(experiment: Dict[str, Any]) -> str:
    """Create a Ludwig config that only contains input and output features.

    Args:
        experiment: experiment dictionary.

    Returns:
        path where the default config is saved.
    """
    model_config = model_configs_for_dataset(experiment["dataset_name"])["default"]

    # only keep input_features and output_features
    main_config_keys = list(model_config.keys())
    for key in main_config_keys:
        if key not in ["input_features", "output_features"]:
            del model_config[key]
    config_path = f"{experiment['dataset_name']}-{uuid.uuid4().hex}.yaml"
    save_yaml(config_path, model_config)
    return config_path


def delete_model_checkpoints(output_directory: str):
    """Deletes outputs of the experiment run that we don't want to save with the artifacts.

    Args:
        output_directory: output directory of the hyperopt run.
    """
    shutil.rmtree(os.path.join(output_directory, "model", "training_checkpoints"), ignore_errors=True)
    if os.path.isfile(os.path.join(output_directory, "model", "model_weights")):
        os.remove(os.path.join(output_directory, "model", "model_weights"))


def delete_hyperopt_outputs(output_directory: str):
    """Deletes outputs of the hyperopt run that we don't want to save with the artifacts.

    Args:
        output_directory: output directory of the hyperopt run.
    """
    for path, currentDirectory, files in os.walk(output_directory):
        for file in files:
            filename = os.path.join(path, file)
            if file not in HYPEROPT_OUTDIR_RETAINED_FILES:
                os.remove(filename)


def save_yaml(filename, dictionary):
    with open(filename, "w") as f:
        yaml.dump(dictionary, f, default_flow_style=False)


def format_time(time_us):
    """Defines how to format time in FunctionEvent.

    from https://github.com/pytorch/pytorch/blob/master/torch/autograd/profiler_util.py
    """
    US_IN_SECOND = 1000.0 * 1000.0
    US_IN_MS = 1000.0
    if time_us >= US_IN_SECOND:
        return f"{time_us / US_IN_SECOND:.3f}s"
    if time_us >= US_IN_MS:
        return f"{time_us / US_IN_MS:.3f}ms"
    return f"{time_us:.3f}us"


def format_memory(nbytes):
    """Returns a formatted memory size string.

    from https://github.com/pytorch/pytorch/blob/master/torch/autograd/profiler_util.py
    """
    KB = 1024
    MB = 1024 * KB
    GB = 1024 * MB
    if abs(nbytes) >= GB:
        return f"{nbytes * 1.0 / GB:.2f} Gb"
    elif abs(nbytes) >= MB:
        return f"{nbytes * 1.0 / MB:.2f} Mb"
    elif abs(nbytes) >= KB:
        return f"{nbytes * 1.0 / KB:.2f} Kb"
    else:
        return str(nbytes) + " b"



from typing import Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BFILL, IMAGE, IMAGENET1K, MISSING_VALUE_STRATEGY_OPTIONS, PREPROCESSING
from ludwig.schema import utils as schema_utils
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import register_preprocessor
from ludwig.schema.metadata import FEATURE_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@register_preprocessor(IMAGE)
@ludwig_dataclass
class ImagePreprocessingConfig(BasePreprocessingConfig):
    missing_value_strategy: str = schema_utils.StringOptions(
        MISSING_VALUE_STRATEGY_OPTIONS,
        default=BFILL,
        allow_none=False,
        description="What strategy to follow when there's a missing value in an image column",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["missing_value_strategy"],
    )

    fill_value: float = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="The maximum number of most common tokens to be considered. If the data contains more than this "
        "amount, the most infrequent tokens will be treated as unknown.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["fill_value"],
    )

    computed_fill_value: float = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="The internally computed fill value to replace missing values with in case the "
        "missing_value_strategy is fill_with_mode or fill_with_mean",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["computed_fill_value"],
    )

    height: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The image height in pixels. If this parameter is set, images will be resized to the specified "
        "height using the resize_method parameter. If None, images will be resized to the size of the "
        "first image in the dataset.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["height"],
    )

    width: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="The image width in pixels. If this parameter is set, images will be resized to the specified "
        "width using the resize_method parameter. If None, images will be resized to the size of the "
        "first image in the dataset.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["width"],
    )

    num_channels: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="Number of channels in the images. If specified, images will be read in the mode specified by the "
        "number of channels. If not specified, the number of channels will be inferred from the image "
        "format of the first valid image in the dataset.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["num_channels"],
    )

    resize_method: str = schema_utils.StringOptions(
        ["crop_or_pad", "interpolate"],
        default="interpolate",
        allow_none=False,
        description="The method to use for resizing images.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["resize_method"],
    )

    infer_image_num_channels: bool = schema_utils.Boolean(
        default=True,
        description="If true, then the number of channels in the dataset is inferred from a sample of the first image "
        "in the dataset.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["infer_image_num_channels"],
    )

    infer_image_dimensions: bool = schema_utils.Boolean(
        default=True,
        description="If true, then the height and width of images in the dataset will be inferred from a sample of "
        "the first image in the dataset. Each image that doesn't conform to these dimensions will be "
        "resized according to resize_method. If set to false, then the height and width of images in the "
        "dataset will be specified by the user.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["infer_image_dimensions"],
    )

    infer_image_max_height: int = schema_utils.PositiveInteger(
        default=256,
        allow_none=False,
        description="If infer_image_dimensions is set, this is used as the maximum height of the images in "
        "the dataset.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["infer_image_max_height"],
    )

    infer_image_max_width: int = schema_utils.PositiveInteger(
        default=256,
        allow_none=False,
        description="If infer_image_dimensions is set, this is used as the maximum width of the images in "
        "the dataset.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["infer_image_max_width"],
    )

    infer_image_sample_size: int = schema_utils.PositiveInteger(
        default=100,
        allow_none=False,
        description="The sample size used for inferring dimensions of images in infer_image_dimensions.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["infer_image_sample_size"],
    )

    standardize_image: Union[str, None] = schema_utils.StringOptions(
        [IMAGENET1K],
        default=None,
        allow_none=True,
        description="Standardize image by per channel mean centering and standard deviation scaling .",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["standardize_image"],
    )

    in_memory: bool = schema_utils.Boolean(
        default=True,
        description="Defines whether image dataset will reside in memory during the training process or will be "
        "dynamically fetched from disk (useful for large datasets). In the latter case a training batch "
        "of input images will be fetched from disk each training iteration.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["in_memory"],
    )

    num_processes: int = schema_utils.PositiveInteger(
        default=1,
        allow_none=False,
        description="Specifies the number of processes to run for preprocessing images.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["num_processes"],
    )

    requires_equal_dimensions: bool = schema_utils.Boolean(
        default=False,
        description="If true, then width and height must be equal.",
        parameter_metadata=FEATURE_METADATA[IMAGE][PREPROCESSING]["requires_equal_dimensions"],
    )
